<?php

namespace App\Service;

use App\Exception;
use App\Exception\IntegrationException as IntegrationException;
use App\Exception\ParamsException as ParamsException;
use App\Exception\NetworkException as NetworkException;
use App\Exception\SqlException as SqlException;


class IntegrationService {
    
    private $util;
    private $productService;
    private $integrationFactory;
    private $branchFactory;
    private $orderFactory;
    private $paymentFactory;

    public function __construct (\App\Util\Util $util, 
                                 \App\Service\ProductService $productService, 
                                 \App\Factory\IntegrationFactory $integrationFactory,
                                 \App\Factory\BranchFactory $branchFactory, 
                                 \App\Factory\OrderFactory $orderFactory, 
                                 \App\Factory\PaymentFactory $paymentFactory) {
        
        $this->util               = $util;
        $this->productService     = $productService;
        $this->integrationFactory = $integrationFactory;
        $this->branchFactory      = $branchFactory;
        $this->orderFactory       = $orderFactory;
        $this->paymentFactory     = $paymentFactory;
    }
    
    public function generalParameters ($branch) {
        $vendor        = $this->branchFactory->getStandartVendor($branch);
        $cashier       = $this->branchFactory->getStandartCashier($branch, $vendor);
        $store         = $this->branchFactory->getStandartStore($branch, $vendor);
        $idControlProd = $this->branchFactory->getProductControlId($branch, $store["CDLOJA"]);
        $state         = $this->branchFactory->getBranchState($branch);
        
        return [
            "CDFILIAL"     => $branch,
            "CDVENDEDOR"   => $vendor,
            "CDLOJA"       => $store["CDLOJA"],
            "CDCAIXA"      => $cashier["CDCAIXA"],
            "CDOPERADOR"   => $cashier["CDOPERADOR"],
            "IDCONTROPROD" => $idControlProd,
            "SGESTADO"     => $state,
            "IMPAUTO"      => (!empty($store["IMPAUTO"]) && $store["IMPAUTO"] != 'N') ? true : false
        ];
    }
    
    public function validateCashierSinc($filial) {
        $caixaSinc = $this->orderFactory->fetchIsCashierSinc($filial);
        if(!isset($caixaSinc) || empty($caixaSinc)){
            throw new ParamsException("Caixa sincronizador não encontrado na filial $filial, Habilite a flag de sincronização de caixa.",404);
        }
    }
    
    public function openOrderBill($generalParams, $order, $consumer, $products) {
        if(empty($generalParams["CDOPERADOR"])) {
            throw new Exception\ParamsException("
                Não existe um operador padrão atrelado ao vendedor padrão {$generalParams['CDVENDEDOR']} da filial {$generalParams['CDFILIAL']}.
                Para atrelar o operador padrão ao vendedor padrão desta filial, verifique a seção 'Problemas Frequentes', no manual da integração.");
        } 
        
        $operator = $this->orderFactory->validateOperator($generalParams);
        if(empty($operator)) {
        throw new Exception\ParamsException("
            O operador {$generalParams['CDOPERADOR']} não está associado a filial {$generalParams['CDFILIAL']}.<br>
            Para atrelar o operador a esta filial, verifique a seção 'Problemas Frequentes', no manual da integração.");
        }
        
        if (!$this->validateDuplication($order["orderNumber"], $order["branch"])){
            throw new IntegrationException("Pedido ". $order["orderNumber"] .", para a filial ". $order["branch"] . " já integrado pelo sistema.");
        }
        
        $sale = $this->util->newCode("VENDAREST".$generalParams["CDFILIAL"], 10);
        $comd = $this->util->newCode("COMANDAVEN".$generalParams["CDFILIAL"], 10);
        
        $order["NRVENDAREST"] = $sale;
        $order["NRCOMANDA"]   = $comd;
        $order["IDSTCOMANDA"] = $generalParams["IDCONTROPROD"] === "S" ? "1" : "3";
        
        if(isset($order["scheduledOrder"]) && !empty($order["scheduledOrder"])) {
            $utcTimezone   = new \DateTimeZone("UTC");
            $orderDateTime = new \DateTime($order["scheduledOrder"]["scheduledDate"], $utcTimezone);
            $storeTimezone = new \DateTimeZone($this->util->getCurrentTimezone($generalParams["SGESTADO"]));
            
            $orderDateTime->setTimezone($storeTimezone);
            
            $order["DTHRAGENDADA"] = date_format($orderDateTime, "d/m/Y H:i:s");
            $order["IDSTCOMANDA"]  = "A";
            $order["IDSINCAGENDA"] = "N";
        } else {
            $order["DTHRAGENDADA"] = null;
            $order["IDSTCOMANDA"]  = $generalParams["IDCONTROPROD"] === "S" ? "1" : "3";
            $order["IDSINCAGENDA"] = "S";
        }
        
        // Observação de pagamento online
        $generalParams['DSOBSCOMANDA'] = "";
        
        foreach($order["total"] as &$total){
            if(isset($total["value"]) && !empty($total["value"])){
                if($total["value"] > 0){
                    $total["value"] = bcdiv($total["value"], "100", 2);
                } else{
                    $total["value"] = "0.00";
                }
            }
        }

        if(!empty($order["paymentMethod"])) {
            $paymentName = !empty($order["paymentMethod"]["cardBrand"]) ? $order["paymentMethod"]["cardBrand"] : $order["paymentMethod"]["name"];
            $str = "{$paymentName}";
            $str .= $order["total"]["requiredChange"]["value"] > "0.00" ? " (VALOR: {$order['total']['requiredChange']['value']}); ": " (VALOR: {$order['total']['total']['value']}); ";
            $generalParams['DSOBSCOMANDA'] .= $str;
        }
        
        if(isset($order["notes"]) && !empty($order["notes"])){
            $generalParams['DSOBSCOMANDA'] .= $order["notes"]; 
        }
        
         /* Fechamento da comanda */
        $this->orderFactory->insertSale($sale, $generalParams, $consumer);
        $this->orderFactory->insertSaleCommand($generalParams, $order, $consumer);
            
        /* Fechamento de itens */
        
        $total = $order["total"]["total"]["value"];

        //MOVCAIXADLV
        $change = "0.00";
        $moneyCode = $this->paymentFactory->getMoneyReceiptCode();
        
        $paymentMethod        = !empty($order["paymentMethod"]["cardBrand"]) ? $order["paymentMethod"]["cardBrand"] : $order["paymentMethod"]["paymentsId"];
        $order["CDTIPORECE"]  = $this->paymentFactory->getReceiptType($paymentMethod, $generalParams["CDFILIAL"]);
        $order["NRSEQMOVDLV"] = $this->util->newCode("MOVCAIXADLV".$generalParams["CDFILIAL"].$sale, 10);
        
        $paymentValue = $total;
        
        if(($order["CDTIPORECE"] == $moneyCode) && $order["total"]["requiredChange"]["value"] != "0.00") {
            $change = bcsub($order["total"]["requiredChange"]["value"], $order["total"]["total"]["value"], 2);
            $paymentValue = $order["total"]["requiredChange"]["value"];
        }
        
        $this->orderFactory->insertCashMoviment($order, $paymentValue, $consumer);

        if($change > "0.00") {
            $movCaixa = $this->util->newCode("MOVCAIXADLV".$generalParams["CDFILIAL"].$sale, 10);
            $this->orderFactory->insertChange($order, $change, $generalParams, $consumer, $movCaixa, $moneyCode);
        }
        
        $generalParams["NRSEQPRODCUP"] = $this->util->newCode("ITCOMANDAVEN".$order["branch"], 10);
        
        $products     = [];
        $nrSeqProdCom = 0;
        
        $ITCOMANDAVEN_SEQ = 0;
        
        foreach ($order["items"] as $item) {
            
            if(empty($item['customCode']) && ($item['price']['value'] != null && $item['price']['value'] != "0.00")){
                throw new Exception\ParamsException("
                        O produto {$item['name']} possui preço e não possui código externo.<br>
                        Para realizar a correção, entre em contato com a equipe de suporte do Delivery Direto.");
            }
            
            if(!empty($item['customCode']) && ($item['price']['value'] == null || $item['price']['value'] == "0.00")) {
                    throw new Exception\ParamsException("
                        O produto {$item['name']} possui preço inferior a R$ 0,01 centavo.<br>
                        Código do produto: {$item['customCode']}.<br>
                        Para realizar a correção, entre em contato com a equipe de suporte do Delivery Direto.");
            }
            
            $processedProduct = $this->productService->processProduct($generalParams, $item, $order, $nrSeqProdCom, $ITCOMANDAVEN_SEQ);
            
            if (is_array($processedProduct) && !isset($processedProduct["CDPRODUTO"])) {
                $products = array_merge($products, $processedProduct);
            } else {
                array_push($products, $processedProduct);
            }
        }
        
        if(empty($products)) 
            throw new \Exception("Ocorreu uma falha na validação de produtos.", 187);
        
        if(bccomp($order["total"]["discount"]["value"], "0", 2) > 0) {
            
            $discountLeft = $this->shareDiscount((string) $order["total"]["discount"]["value"], $products);
            
            $sumItems = bcadd($this->balanceProducts($products), $order["total"]["deliveryFee"]["value"], 2);
            
            if(bccomp($discountLeft, "0", 2) > 0) {
                if(bccomp($order["total"]["deliveryFee"]["value"], $discountLeft, 2) < 0) 
                    throw new IntegrationException("Valor de descontos ( Desconto Total: {$order['total']['discount']['value']} ) superior a somatória dos Itens ( Soma Itens: $sumItems ) <br>{Integração não suporta produtos com 100% de desconto}");
                
                $order["total"]["deliveryFee"]["value"] = bcsub($order["total"]["deliveryFee"]["value"], $discountLeft, 2);
            }
        }

        foreach ($products as $product) {
            
            if(empty($product['VRPRECCOMVEN']) || $product['VRPRECCOMVEN'] < 0) 
                throw new \Exception("O produto ({$product['CDPRODUTO']}) possui preço inferior a 0.01 centavo.");
            
            if($product['VRDESCCOMVEN'] >= $product['VRPRECCOMVEN']) 
                throw new \Exception("O produto ({$product['CDPRODUTO']}) possui DESCONTO superior ao VALOR: Desconto -> {$product['VRDESCCOMVEN']} Valor -> {$product['VRPRECCOMVEN']}", 192);
            
            $this->productService->insertProduct(array_merge($product, $generalParams));
            
            //Observações da OBSITCOMANDAVEN
            if(!empty($product['ITEM_OBSERVATION'])) {
                foreach($product['ITEM_OBSERVATION'] as $itemObservation) {
                    $this->productService->insertObservation($itemObservation);
                }
            }
        }
        
        return [
            'CDFILIAL'      => $generalParams["CDFILIAL"],
            'NRCOMANDA'     => $order["NRCOMANDA"],
            'NRCOMANDAEXT'  => $order["orderNumber"],
            'NRVENDAREST'   => $order["NRVENDAREST"],
            'IMPAUTO'       => $generalParams["IMPAUTO"]
        ];
    }

    public function shareDiscount($initialDiscount, &$products) {

        $idxToApplyDiscount   = [];
        $totalDiscount        = $initialDiscount;
        $discountLeft         = $initialDiscount;
        $productsTotalValue   = "0.00";
        $discountGivenShare   = "0.00";
        $maxValueItem         = "0.00";
        $maxValueIdx          = -1;
        
        foreach($products as &$product) {
            $product['VRDESCCOMVEN']   = 0;
            $product['discountShare']  = "0.00";
            $product['totalItemValue'] = bcmul($product["QTPRODCOMVEN"], $product["VRPRECCOMVEN"], 2);
        }
        
        if(isset($product))
            unset($product);

        foreach($products as $i => $product) {
            if(bccomp($product["VRPRECCOMVEN"], "0.01", 2) == 1) { 
                $productsTotalValue = bcadd($productsTotalValue, $product['totalItemValue'], 2);
                array_push($idxToApplyDiscount, $i);
            }
        }

        foreach($idxToApplyDiscount as $idxWithDiscount) {
            $discountShare = bcdiv((bcmul($product['totalItemValue'], "100.00", 2)), $productsTotalValue, 2);
            $products[$idxWithDiscount]["discountShare"] = $discountShare;

            $discountGivenShare = bcadd($discountGivenShare, $discountShare, 2);
        }

        foreach($products as $i => $product) {
            if(bccomp($product['discountShare'], $maxValueItem, 2) == 1) {
                $maxValueIdx  = $i;
                $maxValueItem = $product['discountShare'];
            }            
        }

        if($maxValueIdx != -1) {
            $products[$maxValueIdx]["discountShare"] = bcadd($products[$maxValueIdx]["discountShare"], bcsub("100.00", $discountGivenShare, 2), 2);
        }
    
        foreach($idxToApplyDiscount as $idxWithDiscount) {
            $itemDiscount = bcdiv((bcmul($products[$idxWithDiscount]["discountShare"], $totalDiscount, 2)), "100.00", 2);

            if(bccomp($products[$idxWithDiscount]["totalItemValue"], $itemDiscount, 2) > 0) {
                $products[$idxWithDiscount]["VRDESCCOMVEN"] = bcdiv($itemDiscount, $products[$idxWithDiscount]["QTPRODCOMVEN"], 2);
                
                $discountLeft = bcsub($discountLeft, $itemDiscount, 2);
            } else if(bccomp($products[$idxWithDiscount]["VRPRECCOMVEN"], $itemDiscount, 2) == 0) {
                $itemDiscount = bcsub($itemDiscount, "0.01", 2);
                $products[$idxWithDiscount]["VRDESCCOMVEN"] = $itemDiscount;
                
                $discountLeft = bcsub($discountLeft, $itemDiscount, 2);
            }
            
        }
        
        if(bccomp($discountLeft, "0", 2) > 0) {
            
            foreach($products as &$p) {
                if(bcsub($p["VRPRECCOMVEN"], bcadd($p["VRDESCCOMVEN"], $discountLeft, 2), 2) > 0) {
                    $p["VRDESCCOMVEN"] = bcadd($p["VRDESCCOMVEN"], $discountLeft, 2);
                    $discountLeft = "0.00";
                }
            }
            
            if (isset($p)) unset($p);
            
        }

        return $discountLeft;
    }
    
    private function balanceProducts($products) {
        return array_reduce($products, function ($carry, $item) {
            return bcadd($carry, $item["VRPRECCOMVEN"], 2);
        }, 0);
    }
    
    public function cancelOrder($order) {
        $orderDetails = $this->orderFactory->getOrderForCancel($order);

        if(empty($orderDetails))
            return 0;        

        if($orderDetails["ORDERSTATUS"] === "X" || $orderDetails["ORDERSTATUS"] === "P" || $orderDetails["ORDERSTATUS"] === "5" || $orderDetails["ORDERSTATUS"] === "E")
            throw new \Exception("Não é possivel cancelar o pedido $order. Nota fiscal já foi expedida");

        $this->orderFactory->cancelOrder($orderDetails);
        $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [Pedido Cancelado] #$order - Pedido cancelado pelo Delivery Direto. \n");
    }
    
    public function validateDuplication($order, $branch){
        return $this->orderFactory->validateDuplication($order, $branch);
    }
    
    public function validateOrder($branch, $sellId){
        $this->orderFactory->validateOrder($branch, $sellId);
    }
    
    public function disableIntegration() {
        $this->integrationFactory->disableIntegration();
    }

    public function getConfiguration() {
        return $this->integrationFactory->getConfiguration();
    }
    
    public function getOrder($orderReference){
        $this->orderFactory->getOrder($orderReference);
    }
}